/********************************************************************
 File: i2c.c

 Description:
 This file contains the functions for I2C communication.

 Authors and Copyright:
 (c) 2018, Thomas Fischl (http://www.fischl.de/contact.html)

 Device: PIC16LF1455
 Compiler: Microchip MPLAB XC8 C Compiler V1.44

 License:
 This file is part of the USBtin firmware project and is copyrighted by the
 authors listed above. It is free for private or educational non-commercial use
 (for a commercial license please contact the authors). It may not be
 redistributed without the prior written consent of the authors.
 
 It is provided "as is" without warranty of any kind, either express or implied,
 including without limitation any implied warranties of condition, uninterrupted
 use, merchantability, fitness for a particular purpose, or non-infringement. In
 no event shall the authors be liable for any direct or indirect damages arising
 in any way out of the use of it.
 
 ********************************************************************/

#include <xc.h>
#include "clock.h"
#include "i2c.h"

unsigned char inprogress = 0;

/**
 * @brief Wait for interrupt flag to be set
 * @retval 1 Flag cleared
 * @retval 0 Timeout
 */
unsigned char waitForIF() {
    unsigned char starttime = clock_tickerSlow;
    while (!SSP1IF) {
        if (clock_diff(starttime) > 5) {
            return 0;
        }
    }
    SSP1IF = 0;
    return 1;
}

/**
 * @brief Initialize I2C unit
 */
void i2c_init() {

    // set baud rate generation
    i2c_setBaudratePeriod(I2C_DEFAULT_PERIOD_US);

    // set I2C master mode
    SSPCON1 = 0b00001000;

    // enable ssp unit
    SSPCON1bits.SSPEN = 1;

}

/**
 * @brief Set baudrate for given period
 * @param period Period of baudrate in us
 */
void i2c_setBaudratePeriod(unsigned short period) {

    if (period == 0) {
        SSPADD = 3; // maximum allowed frequency (0, 1, 2 are not valid values for baudrate generator)
    } else if (period > 21) {
        SSPADD = 0xff; // minimum frequency
    } else {
        SSPADD = (period * (_XTAL_FREQ / 4 / 1000000)) - 1; // BRG = ((Fosc * delay_us) / (4 * 1000000)) - 1
    }
}

/**
 * @brief Send start condition
 * @retval 1 Successful
 * @retval 0 Timeout
 */
unsigned char i2c_start() {

    // mark that we are in progress
    inprogress = 1;
    
    // set start condition
    SSPCON2bits.SEN = 1;
    
    return waitForIF();
}

/**
 * @brief Send repeated start condition
 * @retval 1 Successful
 * @retval 0 Timeout
 */
unsigned char i2c_repeatedStart() {

    // set start condition
    SSPCON2bits.RSEN = 1;

    return waitForIF();
}

/**
 * @brief Send byte to i2c
 *
 * @param b Byte to send out
 * @retval OK Sent byte successfully
 * @retval ERROR Error while sending byte
 */
unsigned char i2c_sendByte(unsigned char b) {

    // send byte
    SSPBUF = b;

    if (!waitForIF()) return 0;

    // Check if there was an ack
    if (SSPCON2bits.ACKSTAT == 0)
        return 1;
    else
        return 0;
}

/**
 * @brief Read byte from I2C
 *
 * @param b Pointer to byte read in from i2c
 * @param ack Acknowledge level which the master should set
 * @retval OK Received byte successfully
 * @retval ERROR Error while receiving byte
 */
unsigned char i2c_receiveByte(unsigned char * b, unsigned char ack) {

    // switch to receiving mode
    SSPCON2bits.RCEN = 1;

    // wait for receive finish	
    if (!waitForIF()) return 0;

    *b = SSPBUF;


    SSPCON2bits.ACKDT = ack;
    SSPCON2bits.ACKEN = 1;

    if (!waitForIF()) return 0;

    return 1;
    /*
        // Check if there was an ack
        if (SSPCON2bits.ACKSTAT == 0)
            return 1;
        else
            return 0;
     */
}

/**
 * @brief Send stop condition
 */
unsigned char i2c_stop(void) {

    // set stop condition
    SSPCON2bits.PEN = 1;

    // mark that we have finished and are free
    inprogress = 0;

    return waitForIF();
}

/**
 * @brief Determine if an I2C operation is currently running
 * @retval 0 No operation running
 * @retval 1 Operation in progress
 */
unsigned char i2c_isInProgress(void) {
    return inprogress;
}