/********************************************************************
 File: main.c

 Description:
 This file contains the main routine with application entry point for
 the USB2FIR firmware project.

 Authors and Copyright:
 (c) 2018, Thomas Fischl (http://www.fischl.de/contact.html)

 Device: PIC16LF1455
 Compiler: Microchip MPLAB XC8 C Compiler V1.44

 License:
 This file is part of the USBtin firmware project and is copyrighted by the
 authors listed above. It is free for private or educational non-commercial use
 (for a commercial license please contact the authors). It may not be
 redistributed without the prior written consent of the authors.
 
 It is provided "as is" without warranty of any kind, either express or implied,
 including without limitation any implied warranties of condition, uninterrupted
 use, merchantability, fitness for a particular purpose, or non-infringement. In
 no event shall the authors be liable for any direct or indirect damages arising
 in any way out of the use of it.
 
 Change History:
 Rev   Date        Description
 1.0   2018-07-10  Initial release
  
 ********************************************************************/

#include <xc.h>
#include "usb.h"
#include "i2c.h"
#include "clock.h"
#include "led.h"
#include "bootloader.h"
#include "mlx_usb.h"
#include "usb2fir.h"


/**
 * @brief Main function and entry point of application
 * @param argc Count of arguments
 * @param argv Argument vector
 * @return Error code
 */
int main(int argc, char** argv) {

    // set clock to 3x PLL, 48 MHz
    OSCCON = 0b11111100;

    // activate active clock tuning with USB clock
    ACTCON = 0b10010000;

    // wait until clock is stable
    while ((OSCSTATbits.HFIOFR == 0) &&
            (OSCSTATbits.PLLRDY == 0));

    ANSELC = 0x00;
    ANSELA = 0x00;

    // set in-/output
    TRISA = 0x00;
    TRISC = 0b00100011;

    // initialize software modules
    clock_init();
    led_init();
    i2c_init();
    usb_init();

    // enable interrupts
    PEIE = 1; // peripheral interrupt enable
    GIE = 1; // enable global interrupts
    
    mlx_usb_init();
    
    
    // main loop
    while (1) {

        // reset / bootloader
        if (!PORTAbits.RA3) {
            usb_shutdown();
            asm("ljmp 0x000");
        }
        
        // do module processing
        usb_process();
        led_process();
        bootloader_process();
        mlx_usb_process();
    };

    return 0;
}

/**
 * @brief Interrupt service routine
 */
void interrupt isr(void) {

    // timer 1 interrupt
    if (TMR1IE && TMR1IF) {
        TMR1IF = 0;

        clock_isr();
    }

}