/********************************************************************
 File: mlx_usb.c

 Description:
 This file contains the USB command processing and data flow handling
 for the USB2FIR firmware project.

 Authors and Copyright:
 (c) 2018, Thomas Fischl (http://www.fischl.de/contact.html)

 Device: PIC16LF1455
 Compiler: Microchip MPLAB XC8 C Compiler V1.44

 License:
 This file is part of the USBtin firmware project and is copyrighted by the
 authors listed above. It is free for private or educational non-commercial use
 (for a commercial license please contact the authors). It may not be
 redistributed without the prior written consent of the authors.
 
 It is provided "as is" without warranty of any kind, either express or implied,
 including without limitation any implied warranties of condition, uninterrupted
 use, merchantability, fitness for a particular purpose, or non-infringement. In
 no event shall the authors be liable for any direct or indirect damages arising
 in any way out of the use of it.
 
 ********************************************************************/


#include "i2c.h"
#include "led.h"
#include "clock.h"
#include "bootloader.h"
#include "usb2fir.h"
#include "usb.h"
#include "mlx_usb.h"

unsigned char status = STATUS_IDLE;
unsigned short expected = 0;
unsigned char saved_cmd;

unsigned short bt_ramaddress = 0;
unsigned char bt_devaddress = MLX_BT_DEVADDRESS_INVALID;
volatile unsigned char * ep1in_buf;


unsigned char mem_access(struct mem_access_cmd *cmd) {
    
    unsigned char address = (cmd->deviceaddress << 1);
        
    i2c_start();
    
    unsigned char success = 1;
    
    if (i2c_sendByte(address) &&
        i2c_sendByte(cmd->memoryaddress >> 8) &&
        i2c_sendByte(cmd->memoryaddress & 0xff)) {    
        
        if (cmd->cmd == CMD_READ_MEMORY) {
            address |= 1;            
            i2c_repeatedStart();
            if (!i2c_sendByte(address)) success = 0;
        }
        
        
    } else {
        success = 0;
    }
    
    if (!success) {
        
        status = STATUS_ADDRESS_NAK;
        expected = 0;
        i2c_stop();
        
    } else {

        status = STATUS_ADDRESS_ACK;
        expected = cmd->length;
        saved_cmd = cmd->cmd;

        /* check if transfer is already done */
        if (!expected)
            i2c_stop();

    }
    
    return (cmd->length ? 0xff : 0x00);
}




/**
 * @brief Handle the USB setup packets
 * 
 * @param data USB packet data for request
 * @param response USB packet data for response
 * @return Length of data to send out
 */
unsigned char usb_vendor_setup(unsigned char * data, unsigned char * response) {

    switch (data[1]) {

        case CMD_ECHO: // echo (for transfer reliability testing)
            response[0] = data[2];
            response[1] = data[3];
            return 2;
            
        case CMD_GET_CAPABILITY:
            response[0] = 0x7f;
            response[1] = 0;
            response[2] = 0;
            response[3] = 0;
            return 4;

        case CMD_START_BOOTLOADER:
            if ((data[2] == 0x37) && (data[3] == 0x52))
                bootloader_triggerStart();
            return 0;
            
        case CMD_GET_STATUS:
            response[0] = status;
            return 1;

        case CMD_READ_MEMORY:
        case CMD_WRITE_MEMORY:
            led_flash();
            return mem_access((struct mem_access_cmd*) data);

        case CMD_CONFIG_BFMODE:
            bt_devaddress = data[2];
            
            // activate step mode
            mlx_writeRegister(bt_devaddress, MLX_REGISTER_CONFIG, 0x1803 | ((data[4] & 0x7) << 7));
            // trigger new measurement (into other subpage)
            mlx_writeRegister(bt_devaddress, MLX_REGISTER_STATUS, 0x0030);
            return 0;
            
    }

    return 0;
}

/**
 * @brief Handle USB in requests
 * 
 * @param data Pointer to packet data to fill
 * @param length Count of maximum data length
 * @return Count of written data
 */
unsigned char usb_vendor_in(unsigned char * data, unsigned char length) {

    unsigned char i;

    if (status == STATUS_ADDRESS_ACK) {

        if (length > expected) {
            length = expected;
        }

        // consume bytes
        for (i = 0; i < length; i++) {
            expected--;
            i2c_receiveByte(&data[i], expected == 0); // TODO: handle return value
        }

        // end transfer on last byte
        if (!expected)
            i2c_stop();

    } else {

        for (i = 0; i < length; i++) {
            data[i] = 0;
        }
    }

    return length;
}

/**
 * @brief Handle USB out requests
 * 
 * @param data Byte field containing USB data packet
 * @param length Length of data packet
 * @return 1 if ack should be sent
 */
unsigned char usb_vendor_out(unsigned char * data, unsigned char length) {

    unsigned char i;
    unsigned char err = 0;

    if (status == STATUS_ADDRESS_ACK) {

        if (length > expected) {
            length = expected;
        }

        // consume bytes
        for (i = 0; i < length; i++) {
            expected--;
            if (!i2c_sendByte(data[i]))
                err = 1;
        }

        // end transfer on last byte
        if (!expected)
            i2c_stop();

        if (err) {
            //TODO: set status
        }

    } else {
        // not in ack state
        return 0;
    }

    return 1;
}

unsigned char mlx_read(unsigned char devaddress, unsigned short regaddress, unsigned short length, unsigned short * data) {
    
    devaddress = devaddress << 1;
    
    i2c_start();
        
    if (!i2c_sendByte(devaddress)) {i2c_stop(); return 0;}
    if (!i2c_sendByte(regaddress >> 8)) {i2c_stop(); return 0;}
    if (!i2c_sendByte(regaddress & 0xff)) {i2c_stop(); return 0;}
    
    i2c_repeatedStart();
    if (!i2c_sendByte(devaddress |= 1)) {i2c_stop(); return 0;}
    
    unsigned char d;
    while (length--) {
        
        i2c_receiveByte(&d, 0); // TODO: handle return value
        *data = d << 8;
        i2c_receiveByte(&d, length == 0); // TODO: handle return value
        *data |= d;
        
        data++;
    }
    
    i2c_stop();
    
    return 1;        
}

unsigned char mlx_readWordToBytefield(unsigned char devaddress, unsigned short regaddress, unsigned char * data) {
    
    devaddress = devaddress << 1;
    
    i2c_start();
        
    if (!i2c_sendByte(devaddress)) {i2c_stop(); return 0;}
    if (!i2c_sendByte(regaddress >> 8)) {i2c_stop(); return 0;}
    if (!i2c_sendByte(regaddress & 0xff)) {i2c_stop(); return 0;}
    
    i2c_repeatedStart();
    if (!i2c_sendByte(devaddress |= 1)) {i2c_stop(); return 0;}
    
    i2c_receiveByte(data, 0); // TODO: handle return value
    data++;
    i2c_receiveByte(data, 1); // TODO: handle return value        
    
    i2c_stop();
    
    return 1;        
}

unsigned char mlx_readChessPattern(unsigned char devaddress, unsigned short ramaddress, unsigned short length, unsigned char * data) {
    
    devaddress = devaddress << 1;
    
    i2c_start();
        
    if (!i2c_sendByte(devaddress)) {i2c_stop(); return 0;}
    if (!i2c_sendByte(ramaddress >> 8)) {i2c_stop(); return 0;}
    if (!i2c_sendByte(ramaddress & 0xff)) {i2c_stop(); return 0;}
    
    i2c_repeatedStart();
    if (!i2c_sendByte(devaddress |= 1)) {i2c_stop(); return 0;}
    
    unsigned char d;
    while (length--) {
        
        i2c_receiveByte(data, 0); // TODO: handle return value
        data++;
        i2c_receiveByte(data, 0); // TODO: handle return value
        data++;        
        i2c_receiveByte(&d, 0); // TODO: handle return value
        i2c_receiveByte(&d, length == 0); // TODO: handle return value
    }
    
    i2c_stop();
    
    return 1;        
}

unsigned char mlx_writeRegister(unsigned char devaddress, unsigned short regaddress, unsigned short value) {
    
    devaddress = devaddress << 1;
    
    i2c_start();
        
    if (!i2c_sendByte(devaddress)) {i2c_stop(); return 0;}
    if (!i2c_sendByte(regaddress >> 8)) {i2c_stop(); return 0;}
    if (!i2c_sendByte(regaddress & 0xff)) {i2c_stop(); return 0;}
    if (!i2c_sendByte(value >> 8)) {i2c_stop(); return 0;}
    if (!i2c_sendByte(value & 0xff)) {i2c_stop(); return 0;}
    
    i2c_stop();
    
    return 1;    
}


void mlx_usb_init() {
    
    ep1in_buf = usb_ep1_getBuffer();
    
}

void mlx_usb_process() {
    
    if ((bt_devaddress == MLX_BT_DEVADDRESS_INVALID) || (!usb_ep1_isAvailable()) || i2c_isInProgress()) return;
    
    if (bt_ramaddress > 0) {

        mlx_readChessPattern(bt_devaddress, bt_ramaddress, 32, (unsigned char*)ep1in_buf);

        bt_ramaddress += 0x40;
        if (bt_ramaddress > MLX_RAM_PIXELDATA_END) {
            bt_ramaddress = 0;
        }

        usb_ep1_arm();            

    } else {

        // read status register
        unsigned short status;
        mlx_read(bt_devaddress, MLX_REGISTER_STATUS, 1, &status);

        if (status & 0x0008) {
            // new data available                                

            ep1in_buf[0] = 0xff;
            ep1in_buf[1] = 0xff;
            ep1in_buf[2] = 0x00;
            ep1in_buf[3] = 0x00;
            ep1in_buf[4] = 0x00;
            ep1in_buf[5] = status & 0x0001;

            mlx_readWordToBytefield(bt_devaddress, 0x700, (unsigned char*)&ep1in_buf[6]); // Ta_Vbe

            if (status & 0x0001)
                mlx_readWordToBytefield(bt_devaddress, 0x728, (unsigned char*)&ep1in_buf[8]); // CP (SP1)
            else 
                mlx_readWordToBytefield(bt_devaddress, 0x708, (unsigned char*)&ep1in_buf[8]); // CP (SP0)

            mlx_readWordToBytefield(bt_devaddress, 0x70A, (unsigned char*)&ep1in_buf[10]); // Gain
            mlx_readWordToBytefield(bt_devaddress, 0x720, (unsigned char*)&ep1in_buf[12]); // Ta_PTAT
            mlx_readWordToBytefield(bt_devaddress, 0x72A, (unsigned char*)&ep1in_buf[14]); // VDDpix

            // pad with zero
            unsigned char i;
            for (i = 16; i < 64; i++) {
                ep1in_buf[i] = 0;
            } 

            // trigger new measurement (into other subpage)
            mlx_writeRegister(bt_devaddress, MLX_REGISTER_STATUS, 0x0030);

            usb_ep1_arm();

            bt_ramaddress = MLX_RAM_PIXELDATA_START + (status & 0x0001);
        }
    }
    
        
}